use crate::{
    admin::{route::Route, station::Station},
    AppState,
};
use axum::Router;
use chrono::{DateTime, Local, NaiveDateTime, Utc};
use serde::{Deserialize, Serialize};
use sqlx::{FromRow, SqlitePool};

mod web;

#[derive(FromRow, Debug, Serialize, Deserialize, PartialEq)]
pub(crate) struct Team {
    pub(crate) id: i64,
    pub(crate) name: String,
    pub(crate) notes: Option<String>,
    pub(crate) amount_people: Option<i64>,
    first_station_id: i64,
    route_id: i64,
}

#[derive(FromRow, Debug, Serialize, Deserialize, PartialEq)]
pub(crate) struct LastContactTeam {
    team_id: i64,
    team_name: String,
    station_id: i64,
    station_name: String,
    last_contact_time: Option<NaiveDateTime>,
}

impl LastContactTeam {
    pub(crate) async fn all_sort_missing(db: &SqlitePool) -> Vec<Self> {
        sqlx::query_as::<_, Self>(
            "SELECT 
    t.id AS team_id,
    t.name AS team_name,
    last_contact.station_id AS station_id,
    s.name AS station_name,
    last_contact.last_contact_time AS last_contact_time
FROM 
    team t
LEFT JOIN (
    SELECT 
        r.team_id,
        r.station_id,
        MAX(COALESCE(r.left_at, r.started_at, r.arrived_at)) AS last_contact_time
    FROM 
        rating r
    GROUP BY 
        r.team_id
    HAVING 
        last_contact_time = MAX(COALESCE(r.left_at, r.started_at, r.arrived_at))
) AS last_contact ON t.id = last_contact.team_id
LEFT JOIN station s ON last_contact.station_id = s.id
ORDER BY 
    last_contact.last_contact_time DESC",
        )
        .fetch_all(db)
        .await
        .unwrap()
    }

    pub(crate) fn local_last_contact(&self) -> Option<DateTime<Local>> {
        let Some(last_contact) = &self.last_contact_time else {
            return None;
        };
        let datetime_utc = DateTime::<Utc>::from_naive_utc_and_offset(*last_contact, Utc);
        Some(datetime_utc.with_timezone(&Local))
    }
}

enum CreateError {
    NoStationForRoute,
    DuplicateName(String),
}

impl Team {
    pub(crate) async fn all(db: &SqlitePool) -> Vec<Self> {
        sqlx::query_as::<_, Self>(
            "SELECT id, name, notes, amount_people, first_station_id, route_id FROM team ORDER BY name;",
        )
        .fetch_all(db)
        .await
        .unwrap()
    }

    pub(crate) async fn all_with_route(db: &SqlitePool, route: &Route) -> Vec<Self> {
        sqlx::query_as!(
            Team,
            "select id, name, notes, amount_people, first_station_id, route_id from team where route_id = ?;",
            route.id
        )
        .fetch_all(db)
        .await
        .unwrap()
    }

    pub(crate) async fn all_with_first_station(db: &SqlitePool, station: &Station) -> Vec<Self> {
        sqlx::query_as!(
            Team,
            "select id, name, notes, amount_people, first_station_id, route_id from team where first_station_id = ?;",
            station.id
        )
        .fetch_all(db)
        .await
        .unwrap()
    }

    pub async fn find_by_id(db: &SqlitePool, id: i64) -> Option<Self> {
        sqlx::query_as!(
            Self,
            "SELECT id, name, notes, amount_people, first_station_id, route_id FROM team WHERE id = ?",
            id
        )
        .fetch_one(db)
        .await
        .ok()
    }

    async fn create(db: &SqlitePool, name: &str, route: &Route) -> Result<i64, CreateError> {
        // get next station id which has the lowest amount of teams to have in the first place
        // assigned
        let Some(station) = route.get_next_first_station(db).await else {
            return Err(CreateError::NoStationForRoute);
        };

        let result = sqlx::query!(
            "INSERT INTO team(name, route_id, first_station_id) VALUES (?, ?, ?) RETURNING id",
            name,
            route.id,
            station.id
        )
        .fetch_one(db)
        .await
        .map_err(|e| CreateError::DuplicateName(e.to_string()))?;
        Ok(result.id.unwrap())
    }

    async fn update_name(&self, db: &SqlitePool, name: &str) {
        sqlx::query!("UPDATE team SET name = ? WHERE id = ?", name, self.id)
            .execute(db)
            .await
            .unwrap();
    }

    async fn update_notes(&self, db: &SqlitePool, notes: &str) {
        sqlx::query!("UPDATE team SET notes = ? WHERE id = ?", notes, self.id)
            .execute(db)
            .await
            .unwrap();
    }

    async fn update_amount_people(&self, db: &SqlitePool, amount_people: i64) {
        sqlx::query!(
            "UPDATE team SET amount_people = ? WHERE id = ?",
            amount_people,
            self.id
        )
        .execute(db)
        .await
        .unwrap();
    }

    async fn update_route(&self, db: &SqlitePool, route: &Route) -> Result<String, ()> {
        let Some(station) = route.get_next_first_station(db).await else {
            return Err(());
        };

        sqlx::query!(
            "UPDATE team SET route_id = ?, first_station_id = ? WHERE id = ?",
            route.id,
            station.id,
            self.id
        )
        .execute(db)
        .await
        .unwrap();

        Ok(station.name)
    }

    async fn update_first_station(&self, db: &SqlitePool, station: &Station) {
        sqlx::query!(
            "UPDATE team SET first_station_id = ? WHERE id = ?",
            station.id,
            self.id
        )
        .execute(db)
        .await
        .unwrap();
    }

    async fn update_amount_people_reset(&self, db: &SqlitePool) {
        sqlx::query!("UPDATE team SET amount_people = NULL WHERE id = ?", self.id)
            .execute(db)
            .await
            .unwrap();
    }

    async fn delete(&self, db: &SqlitePool) -> Result<(), String> {
        sqlx::query!("DELETE FROM team WHERE id = ?", self.id)
            .execute(db)
            .await
            .map_err(|e| e.to_string())?;
        Ok(())
    }

    pub async fn first_station(&self, db: &SqlitePool) -> Station {
        Station::find_by_id(db, self.first_station_id)
            .await
            .expect("db constraints")
    }

    pub async fn route(&self, db: &SqlitePool) -> Route {
        Route::find_by_id(db, self.route_id)
            .await
            .expect("db constraints")
    }

    pub async fn get_curr_points(&self, db: &SqlitePool) -> i64 {
        sqlx::query!(
            "SELECT IFNULL(sum(points), 0) as points FROM rating WHERE team_id = ?",
            self.id
        )
        .fetch_one(db)
        .await
        .unwrap()
        .points
    }
}

pub(super) fn routes() -> Router<AppState> {
    web::routes()
}
